/**
 * DiCloud - GENI project
 * 
 * (C) University of Massachusetts Amherst
 * 
 * Original contributor: Emmanuel Cecchet
 */
package edu.umass.cs.geni.dicloud.handlers;

import java.util.MissingResourceException;
import java.util.Properties;
import java.util.StringTokenizer;

import orca.shirako.plugins.config.Config;
import orca.util.PropList;
import edu.umass.cs.geni.dicloud.AmazonProperties;
import edu.umass.cs.geni.dicloud.NativeCommandExec;

/**
 * This class defines an EBS handler to manage Amazon Elastic Block Storage
 * (EBS) volumes
 * 
 * @author <a href="mailto:cecchet@cs.umass.edu">Emmanuel Cecchet </a>
 * @version 1.0
 * 
 */
public class EBSHandler extends Config
{
  private static final String CREATE_EBS_VOLUME         = "ec2-create-volume";
  private static final String EBS_DELETE_VOLUME_COMMAND = "ec2-delete-volume";
  private static final String EBS_PROBE_COMMAND         = "ec2-describe-volumes";
  private static final String EBS_ATTACH_COMMAND        = "ec2-attach-volume";
  private static final String EBS_DETACH_COMMAND        = "ec2-detach-volume";
  private AWSPlugin           awsPlugin;

  /**
   * Constructor just invoking super
   * 
   * @param plugin
   *          a Shirako plugin
   * 
   */
  public EBSHandler(AWSPlugin plugin)
  {
	super(plugin);
	this.awsPlugin = plugin;
  }

  /**
   * Create a new EBS volume by invoking ec2-create-volume.
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EBS_SIZE: AMI id to use to create instances
   *  - AmazonProperties.EC2_AVAILABILITY_ZONE: availability zone where to create the instance
   * Optional properties: 
   *  - AmazonProperties.EC2_REGION: region where to create the instance
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: command output (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * 
   * AmazonProperties.EBS_VOLUME_ID: Id of the EBS volume
   * AmazonProperties.EBS_CREATION_TIME: Creation time of the volume
   * AmazonProperties.EC2_REGION: region where the volume has been created
   * AmazonProperties.EBS_STATUS: status of the volume
   * AmazonProperties.EBS_CREATION_TIME: volume creation time
   * </pre>
   */
  @Override
  public void join(Object token, Properties p) throws Exception
  {
	// Prepare the command to be executed
	String pk = p.getProperty(AmazonProperties.AWS_PRIVATE_KEY_FILE);
	String cert = p.getProperty(AmazonProperties.AWS_CERT_FILE);
	String region = p.getProperty(AmazonProperties.EC2_REGION);
	String availZone = p.getProperty(AmazonProperties.EC2_AVAILABILITY_ZONE);
	String size = p.getProperty(AmazonProperties.EBS_SIZE_IN_GB);

	// Sanity checks
	if (size == null)
	  throw new MissingResourceException("Missing EBS volume size in join",
		  this.getClass().getName(), AmazonProperties.EBS_SIZE_IN_GB);
	if (pk == null)
	  throw new MissingResourceException("Missing private key file in join",
		  this.getClass().getName(), AmazonProperties.AWS_PRIVATE_KEY_FILE);
	if (cert == null)
	  throw new MissingResourceException(
		  "Missing X509 certificate file in join", this.getClass().getName(),
		  AmazonProperties.AWS_CERT_FILE);
	if (availZone == null)
	  throw new MissingResourceException("Missing availability zone in join",
		  this.getClass().getName(), AmazonProperties.EC2_AVAILABILITY_ZONE);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, TargetJoin);

	// Now create the instances
	if (logger.isInfoEnabled())
	  logger.info("Creating a " + size + " GB EBS Volume");

	String command = CREATE_EBS_VOLUME + " -K " + pk + " -C " + cert + " -s "
	    + size;
	if (region != null)
	  command += " --region " + region;
	if (availZone != null)
	  command += " -z" + availZone;

	// Let's execute the command
	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	if (status == 0)
	{ // Success, check for information about the instance
	  parseOutputAndSetProperties(result, exec);
	}
	else
	{ // Command failed
	  prepareResultError(CREATE_EBS_VOLUME, exec, result);
	}

	returnResult(token, p, result);
  }

  /**
   * Delete an EBS volume by invoking ec2-delete-volume
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EBS_VOLUME_ID: EBS volume id to delete
   * Optional properties: 
   *  - AmazonProperties.EC2_REGION: region where to create the instance
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: command output (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * </pre>
   */
  @Override
  public void leave(Object token, Properties p) throws Exception
  {
	// Prepare the command to be executed
	String volumeId = p.getProperty(AmazonProperties.EBS_VOLUME_ID);
	String pk = p.getProperty(AmazonProperties.AWS_PRIVATE_KEY_FILE);
	String cert = p.getProperty(AmazonProperties.AWS_CERT_FILE);
	String region = p.getProperty(AmazonProperties.EC2_REGION);

	// Sanity checks
	if (volumeId == null)
	  throw new MissingResourceException("Missing volume id in leave", this
		  .getClass().getName(), AmazonProperties.EBS_VOLUME_ID);
	if (pk == null)
	  throw new MissingResourceException("Missing private key file in leave",
		  this.getClass().getName(), AmazonProperties.AWS_PRIVATE_KEY_FILE);
	if (cert == null)
	  throw new MissingResourceException(
		  "Missing X509 certificate file in leave", this.getClass().getName(),
		  AmazonProperties.AWS_CERT_FILE);

	if (logger.isInfoEnabled())
	  logger.info("Deleting volume " + volumeId);

	// Let's execute the command
	String command = EBS_DELETE_VOLUME_COMMAND + " -K " + pk + " -C " + cert;
	if (region != null)
	  command += " --region " + region;
	command += " " + volumeId;

	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, TargetLeave);

	if (status == 0)
	{ // Success, return stdout
	  result.setProperty(Config.PropertyTargetResultCode, Integer
		  .toString(ResultCodeOK));
	  result.setProperty(Config.PropertyTargetResultCodeMessage, exec
		  .getStdout().toString());
	}
	else
	{ // Command failed
	  prepareResultError(EBS_DELETE_VOLUME_COMMAND, exec, result);
	}

	returnResult(token, p, result);
  }

  /**
   * Invoke ec2-describe-volumes to probe the EBS volume. The expected
   * properties are:
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.EC2_AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.EC2_AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EBS_VOLUME_ID: EBS volume id to probe
   * Optional properties: 
   *  - AmazonProperties.EC2_REGION: region where to create the instance
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: AMI status (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * 
   * AmazonProperties.EBS_VOLUME_ID: Id of the EBS volume
   * AmazonProperties.EBS_CREATION_TIME: Creation time of the volume
   * AmazonProperties.EC2_REGION: region where the volume has been created
   * AmazonProperties.EBS_STATUS: status of the volume
   * AmazonProperties.EBS_CREATION_TIME: volume creation time
   * AmazonProperties.EC2_INSTANCE_ID: Instance id volume is attached to (if attached) 
   * AmazonProperties.EBS_DEVICE: Device name under which the volume is know in the instance (if attached)
   * AmazonProperties.EBS_ATTACH_TIME: Time at which the device was attached (if attached)
   * </pre>
   */
  public void probe(Object token, Properties p) throws Exception
  {
	// Prepare the command to be executed
	String volumeId = p.getProperty(AmazonProperties.EBS_VOLUME_ID);
	String pk = p.getProperty(AmazonProperties.AWS_PRIVATE_KEY_FILE);
	String cert = p.getProperty(AmazonProperties.AWS_CERT_FILE);
	String region = p.getProperty(AmazonProperties.EC2_REGION);

	// Sanity checks
	if (volumeId == null)
	  throw new MissingResourceException("Missing volume id in probe", this
		  .getClass().getName(), AmazonProperties.EBS_VOLUME_ID);
	if (pk == null)
	  throw new MissingResourceException("Missing private key file in probe",
		  this.getClass().getName(), AmazonProperties.AWS_PRIVATE_KEY_FILE);
	if (cert == null)
	  throw new MissingResourceException(
		  "Missing X509 certificate file in probe", this.getClass().getName(),
		  AmazonProperties.AWS_CERT_FILE);

	if (logger.isInfoEnabled())
	  logger.info("Probing volume " + volumeId);

	// Let's execute the command
	String command = EBS_PROBE_COMMAND + " -K " + pk + " -C " + cert;
	if (region != null)
	  command += " --region " + region;
	command += " " + volumeId;

	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, AWSPlugin.TargetProbe);

	if (status == 0)
	{ // Success, check for information about the instance
	  parseOutputAndSetProperties(result, exec);
	}
	else
	{ // Command failed
	  prepareResultError(EBS_PROBE_COMMAND, exec, result);
	}

	returnResult(token, p, result);
  }

  /**
   * Invoke ec2-attach-volume to attach an EBS volume to a running instance. The
   * expected properties are:
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.EC2_AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.EC2_AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EBS_VOLUME_ID: EBS volume id to attach
   *  - AmazonProperties.EC2_INSTANCE_ID: Instance to attach to
   *  - AmazonProperties.EBS_DEVICE:  Device name as which to expose VOLUME inside INSTANCE.
   * Optional properties: 
   *  - AmazonProperties.EC2_REGION: region where to create the instance
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: AMI status (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * 
   * AmazonProperties.EBS_STATUS: status of the volume
   * </pre>
   */
  public void attach(Object token, Properties p) throws Exception
  {
	// Prepare the command to be executed
	String volumeId = p.getProperty(AmazonProperties.EBS_VOLUME_ID);
	String pk = p.getProperty(AmazonProperties.AWS_PRIVATE_KEY_FILE);
	String cert = p.getProperty(AmazonProperties.AWS_CERT_FILE);
	String region = p.getProperty(AmazonProperties.EC2_REGION);
	String instanceId = p.getProperty(AmazonProperties.EC2_INSTANCE_ID);
	String device = p.getProperty(AmazonProperties.EBS_DEVICE);

	// Sanity checks
	if (volumeId == null)
	  throw new MissingResourceException("Missing volume id in attach", this
		  .getClass().getName(), AmazonProperties.EBS_VOLUME_ID);
	if (instanceId == null)
	  throw new MissingResourceException("Missing instance id in attach", this
		  .getClass().getName(), AmazonProperties.EC2_INSTANCE_ID);
	if (pk == null)
	  throw new MissingResourceException("Missing private key file in attach",
		  this.getClass().getName(), AmazonProperties.AWS_PRIVATE_KEY_FILE);
	if (cert == null)
	  throw new MissingResourceException(
		  "Missing X509 certificate file in attach", this.getClass().getName(),
		  AmazonProperties.AWS_CERT_FILE);
	if (device == null)
	  throw new MissingResourceException("Missing device name in attach", this
		  .getClass().getName(), AmazonProperties.EBS_DEVICE);

	if (logger.isInfoEnabled())
	  logger.info("Attaching volume " + volumeId + " to instance " + instanceId
		  + " as device " + device);

	// Let's execute the command
	String command = EBS_ATTACH_COMMAND + " -K " + pk + " -C " + cert;
	if (region != null)
	  command += " --region " + region;
	command += " " + volumeId + " -i " + instanceId + " -d " + device;

	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, AWSPlugin.TargetProbe);

	if (status == 0)
	{ // Success, check for information about the instance
	  parseOutputAndSetProperties(result, exec);
	}
	else
	{ // Command failed
	  prepareResultError(EBS_ATTACH_COMMAND, exec, result);
	}

	returnResult(token, p, result);
  }

  /**
   * Invoke ec2-detach-volume to attach an EBS volume to a running instance. The
   * expected properties are:
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.EC2_AWS_PRIVATE_KEY_FILE: AWS account private key
   *  - AmazonProperties.EC2_AWS_CERT_FILE: AWS account X509 certificate
   *  - AmazonProperties.EBS_VOLUME_ID: EBS volume id to attach
   * Optional properties: 
   *  - AmazonProperties.EC2_INSTANCE_ID: Instance to attach to
   *  - AmazonProperties.EBS_DEVICE:  Device name as which to expose VOLUME inside INSTANCE.
   *  - AmazonProperties.EC2_REGION: region where to create the instance
   * </pre>
   * 
   * Returns results in the following properties:
   * 
   * <pre>
   * Config.PropertyTargetResultCode: ResultCodeOK on success, ResultCodeException on error
   * Config.PropertyTargetResultCodeMessage: AMI status (if success)
   * Config.PropertyExceptionMessage: Error message (if error)
   * Config.PropertyExceptionStack: stderr content if command failed
   * 
   * AmazonProperties.EBS_STATUS: status of the volume
   * </pre>
   */
  public void detach(Object token, Properties p) throws Exception
  {
	// Prepare the command to be executed
	String volumeId = p.getProperty(AmazonProperties.EBS_VOLUME_ID);
	String pk = p.getProperty(AmazonProperties.AWS_PRIVATE_KEY_FILE);
	String cert = p.getProperty(AmazonProperties.AWS_CERT_FILE);
	String region = p.getProperty(AmazonProperties.EC2_REGION);
	String instanceId = p.getProperty(AmazonProperties.EC2_INSTANCE_ID);
	String device = p.getProperty(AmazonProperties.EBS_DEVICE);

	// Sanity checks
	if (volumeId == null)
	  throw new MissingResourceException("Missing volume id in detach", this
		  .getClass().getName(), AmazonProperties.EBS_VOLUME_ID);
	if (pk == null)
	  throw new MissingResourceException("Missing private key file in detach",
		  this.getClass().getName(), AmazonProperties.AWS_PRIVATE_KEY_FILE);
	if (cert == null)
	  throw new MissingResourceException(
		  "Missing X509 certificate file in detach", this.getClass().getName(),
		  AmazonProperties.AWS_CERT_FILE);

	if (logger.isInfoEnabled())
	  logger.info("Detaching volume " + volumeId + " from instance "
		  + instanceId + " (known as device " + device + ")");

	// Let's execute the command
	String command = EBS_DETACH_COMMAND + " -K " + pk + " -C " + cert;
	if (region != null)
	  command += " --region " + region;
	command += " " + volumeId;
	if (instanceId != null)
	  command += " -i " + instanceId;
	if (device != null)
	  command += " -d " + device;

	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	// Prepare the result
	Properties result = new Properties();
	result.setProperty(Config.PropertyTargetName, AWSPlugin.TargetProbe);

	if (status == 0)
	{ // Success, check for information about the instance
	  parseOutputAndSetProperties(result, exec);
	}
	else
	{ // Command failed
	  prepareResultError(EBS_DETACH_COMMAND, exec, result);
	}

	returnResult(token, p, result);
  }

  /**
   * Parse the output of the EBS create or describe commands
   * 
   * @param result
   * @param exec
   */
  private void parseOutputAndSetProperties(Properties result,
	  NativeCommandExec exec)
  {
	/**
	 * <pre>
	 * 12:17:04,061 INFO  root Creating a 1 GB EBS Volume
	 * 	Command: ec2-create-volume -K /home/cecchet/ec2/keys/cecchet_pk.pem -C /home/cecchet/ec2/keys/cecchet_cert.pem -s 1 --region us-east-1 -zus-east-1a
	 * 	stdout:[VOLUME	vol-0a1e0863	1		us-east-1a	creating	2010-10-26T16:17:09+0000]
	 * 	stderr:[]
	 * 	Command: ec2-describe-volumes -K /home/cecchet/ec2/keys/cecchet_pk.pem -C /home/cecchet/ec2/keys/cecchet_cert.pem --region us-east-1 vol-0a1e0863
	 * 	stdout:[VOLUME	vol-0a1e0863	1		us-east-1a	available	2010-10-26T16:17:09+0000]
	 * 	stderr:[]
	 * 	12:17:50,405 INFO  root Attaching volume vol-0a1e0863 to instance i-12b1157f as device /dev/hdd
	 * 	Command: ec2-attach-volume -K /home/cecchet/ec2/keys/cecchet_pk.pem -C /home/cecchet/ec2/keys/cecchet_cert.pem --region us-east-1 vol-0a1e0863 -i i-12b1157f -d /dev/hdd
	 * 	stdout:[ATTACHMENT	vol-0a1e0863	i-12b1157f	/dev/hdd	attaching	2010-10-26T16:17:55+0000]
	 * 	stderr:[]
	 * 	12:18:39,392 INFO  root Probing volume vol-0a1e0863
	 * 	Command: ec2-describe-volumes -K /home/cecchet/ec2/keys/cecchet_pk.pem -C /home/cecchet/ec2/keys/cecchet_cert.pem --region us-east-1 vol-0a1e0863
	 * 	stdout:[VOLUME	vol-0a1e0863	1		us-east-1a	in-use	2010-10-26T16:17:09+0000, ATTACHMENT	vol-0a1e0863	i-12b1157f	/dev/hdd	attached	2010-10-26T16:17:55+0000]
	 * 	stderr:[]
	 * </pre>
	 */

	StringTokenizer st = new StringTokenizer(exec.getStdout().toString());
	int fieldNb = 0;
	boolean volume = false;
	while (st.hasMoreTokens())
	{
	  fieldNb++;
	  String field = st.nextToken();
	  // System.out.println("Field[" + fieldNb + "]=" + field);
	  switch (fieldNb)
	  {
	  case 1:
		if ("[VOLUME".equalsIgnoreCase(field))
		  volume = true;
		break;
	  case 2:
		result.setProperty(AmazonProperties.EBS_VOLUME_ID, field);
		break;
	  case 3:
		if (!volume) // instance id
		  result.setProperty(AmazonProperties.EC2_INSTANCE_ID, field);
		else
		  result.remove(AmazonProperties.EC2_INSTANCE_ID);
		// else index
		break;
	  case 4:
		if (volume)
		  result.setProperty(AmazonProperties.EC2_REGION, field);
		else
		  result.setProperty(AmazonProperties.EBS_DEVICE, field);
		break;
	  case 5:
		result.setProperty(AmazonProperties.EBS_STATUS, field);
		break;
	  case 6:
		result.setProperty(AmazonProperties.EBS_CREATION_TIME, field);
		break;
	  // 7: ATTACHMENT
	  // 8: volume id
	  case 9:
		result.setProperty(AmazonProperties.EC2_INSTANCE_ID, field);
		break;
	  case 10:
		result.setProperty(AmazonProperties.EBS_DEVICE, field);
		break;
	  case 11: // Append 'attached' to 'in-use'
		result.setProperty(AmazonProperties.EBS_STATUS, result
		    .getProperty(AmazonProperties.EBS_STATUS)
		    + "/" + field);
		break;
	  case 12:
		result.setProperty(AmazonProperties.EBS_ATTACH_TIME, field);
		break;
	  default:
		break;
	  }
	}

	result.setProperty(Config.PropertyTargetResultCode, Integer
	    .toString(ResultCodeOK));
	result.setProperty(Config.PropertyTargetResultCodeMessage, exec.getStdout()
	    .toString());
  }

  /**
   * Completion of the commands and return of the results
   * 
   * @param token
   * @param p
   * @param result
   */
  private void returnResult(Object token, Properties p, Properties result)
  {
	// pass the configuration sequence number back
	PropList.setProperty(result, Config.PropertyActionSequenceNumber, p
	    .getProperty(Config.PropertyActionSequenceNumber));
	plugin.configurationComplete(token, result);
  }

  /**
   * Add the proper error code and error message to be returned
   * 
   * @param commandName
   *          name of the command executed
   * @param exec
   *          execution object
   * @param result
   *          properties that are returned
   */
  private void prepareResultError(String command, NativeCommandExec exec,
	  Properties result)
  {
	result.setProperty(Config.PropertyTargetResultCode, Integer
	    .toString(ResultCodeException));
	result.setProperty(PropertyExceptionMessage, command + " execution failed");
	result.setProperty(PropertyExceptionStack, exec.getStderr().toString());
  }

  /**
   * Debug the output of a command being executed
   * 
   * @param commandName
   *          name of the command executed
   * @param status
   *          status returned by the command execution
   * @param exec
   *          execution object
   */
  private void debugCommandOutput(String commandName, int status,
	  NativeCommandExec exec)
  {
	if (logger.isDebugEnabled())
	{
	  logger.debug(commandName + " status: " + status);
	  logger.debug("stdout:" + exec.getStdout());
	  logger.debug("stderr:" + exec.getStderr());
	}

	System.err.println("Command: " + commandName);
	System.out.println("stdout:" + exec.getStdout());
	System.out.println("stderr:" + exec.getStderr());
  }

}
