/**
 * DiCloud - GENI project
 * 
 * (C) 2010 - University of Massachusetts Amherst
 * 
 * Initial developer(s): Emmanuel Cecchet.
 * Contributor(s): ______________________.
 */

package edu.umass.cs.geni.dicloud.server;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.MissingResourceException;
import java.util.Properties;
import java.util.StringTokenizer;

import org.apache.log4j.Logger;

import edu.umass.cs.geni.dicloud.AmazonProperties;
import edu.umass.cs.geni.dicloud.NativeCommandExec;
import edu.umass.cs.geni.dicloud.accounting.AWSActivity;
import edu.umass.cs.geni.dicloud.accounting.AWSActivityPricing;
import edu.umass.cs.geni.dicloud.handlers.AWSPlugin;

/**
 * S3 proxy for put/get operations
 * 
 * @author <a href="mailto:cecchet@cs.umass.edu">Emmanuel Cecchet </a>
 * @version 1.0
 * 
 */
public class S3Proxy extends UnicastRemoteObject implements S3ProxyInterface
{
  private static final long   serialVersionUID = -1719815564497600468L;
  private static final double ONE_GB           = (1024 * 1024 * 1024);
  private static final String S3_PUT_COMMAND   = "s3cmd put";
  private static final String S3_GET_COMMAND   = "s3cmd get";
  private static final String S3_DU_COMMAND    = "s3cmd du";

  private String              bucketName;
  private String              region;
  private Logger              logger           = Logger.getRootLogger();
  private AWSPlugin           awsPlugin;

  /**
   * Create a new S3 proxy for a given bucket in a given region.
   * 
   * <pre>
   * Mandatory properties:
   *  - AmazonProperties.S3_BUCKET_NAME: name of the S3 bucket
   * Optional properties: 
   *  - AmazonProperties.S3_LOCATION: region where the bucket resides
   * </pre>
   * 
   * @param plugin
   *          the AWS plugin
   * @param p
   *          Properties as defined above
   */
  public S3Proxy(AWSPlugin plugin, Properties p) throws RemoteException
  {
	// Prepare the command to be executed
	this.bucketName = p.getProperty(AmazonProperties.S3_BUCKET_NAME);
	this.region = p.getProperty(AmazonProperties.S3_LOCATION);

	// Sanity checks
	if (bucketName == null)
	  throw new MissingResourceException("Missing S3 bucket name in S3 Proxy",
		  this.getClass().getName(), AmazonProperties.S3_BUCKET_NAME);

	this.awsPlugin = plugin;
  }

  /*
   * (non-Javadoc)
   * 
   * @see
   * edu.umass.cs.geni.dicloud.accounting.S3ProxyInterface#put(java.lang.String,
   * byte[])
   */
  public void put(String objectName, byte[] data) throws Exception,
	  RemoteException
  {
	if (!awsPlugin.getAwsAccouting().hasCreditLeft())
	  throw new Exception("No AWS credit left to allocate resources");

	long bucketSize = getCurrentBucketSize();

	// Create a file named objectName and put data in it
	if (logger.isInfoEnabled())
	  logger.info("Creating file " + objectName);

	File f = new File(objectName);
	f.createNewFile(); // if a file with that name already exists, it will fail
	// here
	FileOutputStream fos = new FileOutputStream(f);
	fos.write(data);
	fos.close();

	// Now execute 's3cmd put'
	String command = S3_PUT_COMMAND + " ";
	if (region != null)
	  command += "--bucket-location=" + region + " ";
	command += objectName + " " + "s3://" + bucketName;
	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);
	debugCommandOutput(command, status, exec);

	// Delete the temporary file
	f.delete();

	if (status == 0)
	{
	  /*
	   * Account for the PUT operation and the 2 LIST operations to get the
	   * bucket size.
	   */
	  awsPlugin.getAwsAccouting().logActivity(
		  new AWSActivity("S3", command,
		      AWSActivityPricing.S3_PUT_OPERATION * 3, System
		          .currentTimeMillis()));

	  // Calculate S3 storage space
	  long newSize = getCurrentBucketSize();
	  long additionalGB = (long) ((newSize / (long) ONE_GB) - (bucketSize / (long) ONE_GB));
	  if (additionalGB > 0)
	  {
		logger.info(additionalGB + " GB of additional S3 storage allocated");
		awsPlugin.getAwsAccouting().logActivity(
		    new AWSActivity("S3", "storage",
		        AWSActivityPricing.S3_STORAGE_PER_GB_PER_MONTH * additionalGB,
		        System.currentTimeMillis()));
	  }

	  // Log Network In traffic
	  awsPlugin.getAwsAccouting().logActivity(
		  new AWSActivity("S3", "Network In traffic for " + command,
		      AWSActivityPricing.EC2_NETWORK_IN_PER_GB * data.length / ONE_GB,
		      System.currentTimeMillis()));

	  return; // Success
	}
	else
	  // Command failed
	  throw new RuntimeException(command + " execution failed: "
		  + exec.getStderr().toString());
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   */
  private long getCurrentBucketSize() throws IOException, InterruptedException
  {
	// Retrieve the current bucket size
	String command = S3_DU_COMMAND + " ";
	if (region != null)
	  command += "--bucket-location=" + region + " ";
	command += "s3://" + bucketName;
	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);
	debugCommandOutput(command, status, exec);
	if (status == 0)
	{
	  /*
	   * Sample output: $ s3cmd du s3://geni-dicloud-s3handlertest 19722993
	   * s3://geni-dicloud-s3handlertest/
	   */
	  return Long.valueOf(
		  new StringTokenizer(exec.getStdout().get(0)).nextToken()).longValue();
	}
	else
	  // Command failed
	  throw new RuntimeException("Unable to retrieve bucket size using "
		  + command + ": " + exec.getStderr().toString());
  }

  /*
   * (non-Javadoc)
   * 
   * @see
   * edu.umass.cs.geni.dicloud.accounting.S3ProxyInterface#get(java.lang.String)
   */
  public byte[] get(String objectName) throws RemoteException, Exception
  {
	if (!awsPlugin.getAwsAccouting().hasCreditLeft())
	  throw new Exception("No AWS credit left to allocate resources");

	File f = File.createTempFile("dicloud", "get");

	// Execute 's3cmd get'
	String command = S3_GET_COMMAND + " -f ";
	if (region != null)
	  command += "--bucket-location=" + region + " ";
	command += "s3://" + bucketName + "/" + objectName + " " + f.getPath();

	// Let's execute the command
	NativeCommandExec exec = new NativeCommandExec();
	int status = exec.executeNativeCommand(command, null, null, 0, false);

	debugCommandOutput(command, status, exec);

	if (status == 0)
	{
	  // Retrieve data from the file
	  if (logger.isDebugEnabled())
		logger.debug("Opening file " + objectName);

	  int size = (int) f.length();
	  byte b[] = new byte[size];
	  FileInputStream fis = new FileInputStream(f);
	  fis.read(b);
	  fis.close();
	  // Delete the temporary file
	  f.delete();

	  // Account for network transfer
	  awsPlugin.getAwsAccouting().logActivity(
		  new AWSActivity("S3", "Network Out traffic for " + command,
		      AWSActivityPricing.EC2_NETWORK_OUT_PER_GB * (size / ONE_GB),
		      System.currentTimeMillis()));

	  // Log GET operation
	  awsPlugin.getAwsAccouting().logActivity(
		  new AWSActivity("S3", command, AWSActivityPricing.S3_GET_OPERATION,
		      System.currentTimeMillis()));

	  return b; // Success
	}
	else
	  // Command failed
	  throw new RuntimeException(command + " execution failed: "
		  + exec.getStderr().toString());
  }

  /**
   * Debug the output of a command being executed
   * 
   * @param commandName
   *          name of the command executed
   * @param status
   *          status returned by the command execution
   * @param exec
   *          execution object
   */
  private void debugCommandOutput(String commandName, int status,
	  NativeCommandExec exec)
  {
	if (logger.isDebugEnabled())
	{
	  logger.debug(commandName + " status: " + status);
	  logger.debug("stdout:" + exec.getStdout());
	  logger.debug("stderr:" + exec.getStderr());
	}
	if (status != 0)
	  logger.warn("Command " + commandName + " failed with error: "
		  + exec.getStderr());
  }

}
